#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

import json
import logging
import uuid
from dataclasses import InitVar, dataclass, field
from pathlib import Path
from typing import Any, Mapping, Optional, Union

from airbyte_cdk.models import AirbyteRecordMessageFileReference
from airbyte_cdk.sources.declarative.extractors.record_extractor import RecordExtractor
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import (
    InterpolatedString,
)
from airbyte_cdk.sources.declarative.partition_routers.substream_partition_router import (
    SafeResponse,
)
from airbyte_cdk.sources.declarative.requesters import Requester
from airbyte_cdk.sources.declarative.types import Record, StreamSlice
from airbyte_cdk.sources.types import Config
from airbyte_cdk.sources.utils.files_directory import get_files_directory

logger = logging.getLogger("airbyte")


@dataclass
class FileUploader:
    requester: Requester
    download_target_extractor: RecordExtractor
    config: Config
    parameters: InitVar[Mapping[str, Any]]

    filename_extractor: Optional[Union[InterpolatedString, str]] = None
    content_extractor: Optional[RecordExtractor] = None

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        if self.filename_extractor:
            self.filename_extractor = InterpolatedString.create(
                self.filename_extractor,
                parameters=parameters,
            )

    def upload(self, record: Record) -> None:
        mocked_response = SafeResponse()
        mocked_response.content = json.dumps(record.data).encode()
        download_targets = list(self.download_target_extractor.extract_records(mocked_response))
        if not download_targets:
            raise ValueError("No download targets found")

        download_target = download_targets[0]  # we just expect one download target
        if not isinstance(download_target, str):
            raise ValueError(
                f"download_target is expected to be a str but was {type(download_target)}: {download_target}"
            )

        response = self.requester.send_request(
            stream_slice=StreamSlice(
                partition={}, cursor_slice={}, extra_fields={"download_target": download_target}
            ),
        )

        if self.content_extractor:
            raise NotImplementedError("TODO")
        else:
            files_directory = Path(get_files_directory())

            file_name = (
                self.filename_extractor.eval(self.config, record=record)
                if self.filename_extractor
                else str(uuid.uuid4())
            )
            file_name = file_name.lstrip("/")
            file_relative_path = Path(record.stream_name) / Path(file_name)

            full_path = files_directory / file_relative_path
            full_path.parent.mkdir(parents=True, exist_ok=True)

            with open(str(full_path), "wb") as f:
                f.write(response.content)
            file_size_bytes = full_path.stat().st_size

            logger.info("File uploaded successfully")
            logger.info(f"File url: {str(full_path)}")
            logger.info(f"File size: {file_size_bytes / 1024} KB")
            logger.info(f"File relative path: {str(file_relative_path)}")

            record.file_reference = AirbyteRecordMessageFileReference(
                staging_file_url=str(full_path),
                source_file_relative_path=str(file_relative_path),
                file_size_bytes=file_size_bytes,
            )



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

import time
from typing import Any, Optional

from airbyte_cdk.models import (
    AirbyteAnalyticsTraceMessage,
    AirbyteMessage,
    AirbyteTraceMessage,
    TraceType,
    Type,
)


def create_analytics_message(type: str, value: Optional[Any]) -> AirbyteMessage:
    return AirbyteMessage(
        type=Type.TRACE,
        trace=AirbyteTraceMessage(
            type=TraceType.ANALYTICS,
            emitted_at=time.time() * 1000,
            analytics=AirbyteAnalyticsTraceMessage(
                type=type, value=str(value) if value is not None else None
            ),
        ),
    )



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#
import os

AIRBYTE_STAGING_DIRECTORY = os.getenv("AIRBYTE_STAGING_DIRECTORY", "/staging/files")
DEFAULT_LOCAL_DIRECTORY = "/tmp/airbyte-file-transfer"


def get_files_directory() -> str:
    return (
        AIRBYTE_STAGING_DIRECTORY
        if os.path.exists(AIRBYTE_STAGING_DIRECTORY)
        else DEFAULT_LOCAL_DIRECTORY
    )



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.

from dataclasses import InitVar, dataclass
from typing import Any, Iterable, List, Mapping, Optional, Union

from airbyte_cdk.sources.declarative.requesters.query_properties import (
    PropertiesFromEndpoint,
    PropertyChunking,
)
from airbyte_cdk.sources.types import Config, StreamSlice


@dataclass
class QueryProperties:
    """
    Low-code component that encompasses the behavior to inject additional property values into the outbound API
    requests. Property values can be defined statically within the manifest or dynamically by making requests
    to a partner API to retrieve the properties. Query properties also allow for splitting of the total set of
    properties into smaller chunks to satisfy API restrictions around the total amount of data retrieved
    """

    property_list: Optional[Union[List[str], PropertiesFromEndpoint]]
    always_include_properties: Optional[List[str]]
    property_chunking: Optional[PropertyChunking]
    config: Config
    parameters: InitVar[Mapping[str, Any]]

    def get_request_property_chunks(
        self, stream_slice: Optional[StreamSlice] = None
    ) -> Iterable[List[str]]:
        """
        Uses the defined property_list to fetch the total set of properties dynamically or from a static list
        and based on the resulting properties, performs property chunking if applicable.
        :param stream_slice: The StreamSlice of the current partition being processed during the sync. This is included
        because subcomponents of QueryProperties can make use of interpolation of the top-level StreamSlice object
        """
        fields: Union[Iterable[str], List[str]]
        if isinstance(self.property_list, PropertiesFromEndpoint):
            fields = self.property_list.get_properties_from_endpoint(stream_slice=stream_slice)
        else:
            fields = self.property_list if self.property_list else []

        if self.property_chunking:
            yield from self.property_chunking.get_request_property_chunks(
                property_fields=fields, always_include_properties=self.always_include_properties
            )
        else:
            yield list(fields)

    # delete later, but leaving this to keep the discussion thread on the PR from getting hidden
    def has_multiple_chunks(self, stream_slice: Optional[StreamSlice]) -> bool:
        property_chunks = iter(self.get_request_property_chunks(stream_slice=stream_slice))
        try:
            next(property_chunks)
            next(property_chunks)
            return True
        except StopIteration:
            return False



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass, field
from typing import Any, List, Mapping, MutableMapping, Optional, Union

import requests

from airbyte_cdk.sources.declarative.requesters.error_handlers.default_http_response_filter import (
    DefaultHttpResponseFilter,
)
from airbyte_cdk.sources.declarative.requesters.error_handlers.http_response_filter import (
    HttpResponseFilter,
)
from airbyte_cdk.sources.streams.http.error_handlers import BackoffStrategy, ErrorHandler
from airbyte_cdk.sources.streams.http.error_handlers.response_models import (
    SUCCESS_RESOLUTION,
    ErrorResolution,
    create_fallback_error_resolution,
)
from airbyte_cdk.sources.types import Config


@dataclass
class DefaultErrorHandler(ErrorHandler):
    """
    Default error handler.

    By default, the handler will only use the `DEFAULT_ERROR_MAPPING` that is part of the Python CDK's `HttpStatusErrorHandler`.

    If the response is successful, then a SUCCESS_RESOLUTION is returned.
    Otherwise, iterate over the response_filters.
    If any of the filter match the response, then return the appropriate status.
    When `DefaultErrorHandler.backoff_time()` is invoked, iterate sequentially over the backoff_strategies and return the first non-None backoff time, else return None.

    Sample configs:

    1. retry 10 times
    `
        error_handler:
          max_retries: 10
    `
    2. backoff for 5 seconds
    `
        error_handler:
          backoff_strategies:
            - type: "ConstantBackoff"
              backoff_time_in_seconds: 5
    `
    3. retry on HTTP 404
    `
        error_handler:
          response_filters:
            - http_codes: [ 404 ]
              action: RETRY
    `
    4. ignore HTTP 404
    `
      error_handler:
        response_filters:
          - http_codes: [ 404 ]
            action: IGNORE
    `
    5. retry if error message contains `retrythisrequest!` substring
    `
        error_handler:
          response_filters:
            - error_message_contain: "retrythisrequest!"
              action: IGNORE
    `
    6. retry if 'code' is a field present in the response body
    `
        error_handler:
          response_filters:
            - predicate: "{{ 'code' in response }}"
              action: IGNORE
    `

    7. ignore 429 and retry on 404
    `
        error_handler:
        - http_codes: [ 429 ]
          action: IGNORE
        - http_codes: [ 404 ]
          action: RETRY
    `

    Attributes:
        response_filters (Optional[List[HttpResponseFilter]]): response filters to iterate on
        max_retries (Optional[int]): maximum retry attempts
        backoff_strategies (Optional[List[BackoffStrategy]]): list of backoff strategies to use to determine how long
        to wait before retrying
    """

    parameters: InitVar[Mapping[str, Any]]
    config: Config
    response_filters: Optional[List[HttpResponseFilter]] = None
    max_retries: Optional[int] = 5
    max_time: int = 60 * 10
    _max_retries: int = field(init=False, repr=False, default=5)
    _max_time: int = field(init=False, repr=False, default=60 * 10)
    backoff_strategies: Optional[List[BackoffStrategy]] = None

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        if not self.response_filters:
            self.response_filters = [HttpResponseFilter(config=self.config, parameters={})]

        self._last_request_to_attempt_count: MutableMapping[requests.PreparedRequest, int] = {}

    def interpret_response(
        self, response_or_exception: Optional[Union[requests.Response, Exception]]
    ) -> ErrorResolution:
        if self.response_filters:
            for response_filter in self.response_filters:
                matched_error_resolution = response_filter.matches(
                    response_or_exception=response_or_exception
                )
                if matched_error_resolution:
                    return matched_error_resolution
        if isinstance(response_or_exception, requests.Response):
            if response_or_exception.ok:
                return SUCCESS_RESOLUTION

        default_reponse_filter = DefaultHttpResponseFilter(parameters={}, config=self.config)
        default_response_filter_resolution = default_reponse_filter.matches(response_or_exception)

        return (
            default_response_filter_resolution
            if default_response_filter_resolution
            else create_fallback_error_resolution(response_or_exception)
        )

    def backoff_time(
        self,
        response_or_exception: Optional[Union[requests.Response, requests.RequestException]],
        attempt_count: int = 0,
    ) -> Optional[float]:
        backoff = None
        if self.backoff_strategies:
            for backoff_strategy in self.backoff_strategies:
                backoff = backoff_strategy.backoff_time(
                    response_or_exception=response_or_exception, attempt_count=attempt_count
                )
                if backoff:
                    return backoff
        return backoff



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Base class for connector test suites."""

from __future__ import annotations

import abc
import inspect
import sys
from collections.abc import Callable
from pathlib import Path
from typing import cast

import yaml
from boltons.typeutils import classproperty

from airbyte_cdk.models import (
    AirbyteMessage,
    Type,
)
from airbyte_cdk.test import entrypoint_wrapper
from airbyte_cdk.test.standard_tests._job_runner import IConnector, run_test_job
from airbyte_cdk.test.standard_tests.models import (
    ConnectorTestScenario,
)

ACCEPTANCE_TEST_CONFIG = "acceptance-test-config.yml"
MANIFEST_YAML = "manifest.yaml"


class ConnectorTestSuiteBase(abc.ABC):
    """Base class for connector test suites."""

    connector: type[IConnector] | Callable[[], IConnector] | None = None
    """The connector class or a factory function that returns an scenario of IConnector."""

    @classmethod
    def get_test_class_dir(cls) -> Path:
        """Get the file path that contains the class."""
        module = sys.modules[cls.__module__]
        # Get the directory containing the test file
        return Path(inspect.getfile(module)).parent

    @classmethod
    def create_connector(
        cls,
        scenario: ConnectorTestScenario,
    ) -> IConnector:
        """Instantiate the connector class."""
        connector = cls.connector  # type: ignore
        if connector:
            if callable(connector) or isinstance(connector, type):
                # If the connector is a class or factory function, instantiate it:
                return cast(IConnector, connector())  # type: ignore [redundant-cast]

        # Otherwise, we can't instantiate the connector. Fail with a clear error message.
        raise NotImplementedError(
            "No connector class or connector factory function provided. "
            "Please provide a class or factory function in `cls.connector`, or "
            "override `cls.create_connector()` to define a custom initialization process."
        )

    # Test Definitions

    def test_check(
        self,
        scenario: ConnectorTestScenario,
    ) -> None:
        """Run `connection` acceptance tests."""
        result: entrypoint_wrapper.EntrypointOutput = run_test_job(
            self.create_connector(scenario),
            "check",
            test_scenario=scenario,
        )
        conn_status_messages: list[AirbyteMessage] = [
            msg for msg in result._messages if msg.type == Type.CONNECTION_STATUS
        ]  # noqa: SLF001  # Non-public API
        assert len(conn_status_messages) == 1, (
            f"Expected exactly one CONNECTION_STATUS message. Got: {result._messages}"
        )

    @classmethod
    def get_connector_root_dir(cls) -> Path:
        """Get the root directory of the connector."""
        for parent in cls.get_test_class_dir().parents:
            if (parent / MANIFEST_YAML).exists():
                return parent
            if (parent / ACCEPTANCE_TEST_CONFIG).exists():
                return parent
            if parent.name == "airbyte_cdk":
                break
        # If we reach here, we didn't find the manifest file in any parent directory
        # Check if the manifest file exists in the current directory
        for parent in Path.cwd().parents:
            if (parent / MANIFEST_YAML).exists():
                return parent
            if (parent / ACCEPTANCE_TEST_CONFIG).exists():
                return parent
            if parent.name == "airbyte_cdk":
                break

        raise FileNotFoundError(
            "Could not find connector root directory relative to "
            f"'{str(cls.get_test_class_dir())}' or '{str(Path.cwd())}'."
        )

    @classproperty
    def acceptance_test_config_path(cls) -> Path:
        """Get the path to the acceptance test config file."""
        result = cls.get_connector_root_dir() / ACCEPTANCE_TEST_CONFIG
        if result.exists():
            return result

        raise FileNotFoundError(f"Acceptance test config file not found at: {str(result)}")

    @classmethod
    def get_scenarios(
        cls,
    ) -> list[ConnectorTestScenario]:
        """Get acceptance tests for a given category.

        This has to be a separate function because pytest does not allow
        parametrization of fixtures with arguments from the test class itself.
        """
        category = "connection"
        all_tests_config = yaml.safe_load(cls.acceptance_test_config_path.read_text())
        if "acceptance_tests" not in all_tests_config:
            raise ValueError(
                f"Acceptance tests config not found in {cls.acceptance_test_config_path}."
                f" Found only: {str(all_tests_config)}."
            )
        if category not in all_tests_config["acceptance_tests"]:
            return []
        if "tests" not in all_tests_config["acceptance_tests"][category]:
            raise ValueError(f"No tests found for category {category}")

        tests_scenarios = [
            ConnectorTestScenario.model_validate(test)
            for test in all_tests_config["acceptance_tests"][category]["tests"]
            if "iam_role" not in test["config_path"]
        ]
        connector_root = cls.get_connector_root_dir().absolute()
        for test in tests_scenarios:
            if test.config_path:
                test.config_path = connector_root / test.config_path
            if test.configured_catalog_path:
                test.configured_catalog_path = connector_root / test.configured_catalog_path

        return tests_scenarios



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

from typing import Mapping

from pydantic.v1 import BaseModel

from airbyte_cdk.sources.declarative.checks.check_dynamic_stream import CheckDynamicStream
from airbyte_cdk.sources.declarative.checks.check_stream import (
    CheckStream,
    DynamicStreamCheckConfig,
)
from airbyte_cdk.sources.declarative.checks.connection_checker import ConnectionChecker
from airbyte_cdk.sources.declarative.models import (
    CheckDynamicStream as CheckDynamicStreamModel,
)
from airbyte_cdk.sources.declarative.models import (
    CheckStream as CheckStreamModel,
)

COMPONENTS_CHECKER_TYPE_MAPPING: Mapping[str, type[BaseModel]] = {
    "CheckStream": CheckStreamModel,
    "CheckDynamicStream": CheckDynamicStreamModel,
}

__all__ = ["CheckStream", "CheckDynamicStream", "ConnectionChecker", "DynamicStreamCheckConfig"]



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Mapping


@dataclass
class TypeTransformer(ABC):
    """
    Abstract base class for implementing type transformation logic.

    This class provides a blueprint for defining custom transformations
    on data records based on a provided schema. Implementing classes
    must override the `transform` method to specify the transformation
    logic.

    Attributes:
        None explicitly defined, as this is a dataclass intended to be
        subclassed.

    Methods:
        transform(record: Dict[str, Any], schema: Mapping[str, Any]) -> None:
            Abstract method that must be implemented by subclasses.
            It performs a transformation on a given data record based
            on the provided schema.

    Usage:
        To use this class, create a subclass that implements the
        `transform` method with the desired transformation logic.
    """

    @abstractmethod
    def transform(
        self,
        record: Dict[str, Any],
        schema: Mapping[str, Any],
    ) -> None:
        """
        Perform a transformation on a data record based on a given schema.

        Args:
            record (Dict[str, Any]): The data record to be transformed.
            schema (Mapping[str, Any]): The schema that dictates how
                the record should be transformed.

        Returns:
            None

        Raises:
            NotImplementedError: If the method is not implemented
                by a subclass.
        """



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

from datetime import datetime
from typing import Optional

from pydantic.v1 import BaseModel


class FileRecordData(BaseModel):
    """
    A record in a file-based stream.
    """

    folder: str
    filename: str
    bytes: int
    source_uri: str
    id: Optional[str] = None
    created_at: Optional[str] = None
    updated_at: Optional[str] = None
    mime_type: Optional[str] = None



#
# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
#

from typing import Optional, Union

import requests

from airbyte_cdk.sources.declarative.requesters.error_handlers.http_response_filter import (
    HttpResponseFilter,
)
from airbyte_cdk.sources.streams.http.error_handlers.default_error_mapping import (
    DEFAULT_ERROR_MAPPING,
)
from airbyte_cdk.sources.streams.http.error_handlers.response_models import (
    ErrorResolution,
    create_fallback_error_resolution,
)


class DefaultHttpResponseFilter(HttpResponseFilter):
    def matches(
        self, response_or_exception: Optional[Union[requests.Response, Exception]]
    ) -> Optional[ErrorResolution]:
        default_mapped_error_resolution = None

        if isinstance(response_or_exception, (requests.Response, Exception)):
            mapped_key: Union[int, type] = (
                response_or_exception.status_code
                if isinstance(response_or_exception, requests.Response)
                else response_or_exception.__class__
            )

            default_mapped_error_resolution = DEFAULT_ERROR_MAPPING.get(mapped_key)

        return (
            default_mapped_error_resolution
            if default_mapped_error_resolution
            else create_fallback_error_resolution(response_or_exception)
        )



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from datetime import timedelta
from typing import Any, List, Mapping, Optional, Sequence, Tuple, Union

import dpath

from airbyte_cdk.config_observation import (
    create_connector_config_control_message,
    emit_configuration_as_airbyte_control_message,
)
from airbyte_cdk.sources.message import MessageRepository, NoopMessageRepository
from airbyte_cdk.sources.streams.http.requests_native_auth.abstract_oauth import (
    AbstractOauth2Authenticator,
)
from airbyte_cdk.utils.datetime_helpers import (
    AirbyteDateTime,
    ab_datetime_now,
    ab_datetime_parse,
)


class Oauth2Authenticator(AbstractOauth2Authenticator):
    """
    Generates OAuth2.0 access tokens from an OAuth2.0 refresh token and client credentials.
    The generated access token is attached to each request via the Authorization header.
    If a connector_config is provided any mutation of it's value in the scope of this class will emit AirbyteControlConnectorConfigMessage.
    """

    def __init__(
        self,
        token_refresh_endpoint: str,
        client_id: str,
        client_secret: str,
        refresh_token: str,
        client_id_name: str = "client_id",
        client_secret_name: str = "client_secret",
        refresh_token_name: str = "refresh_token",
        scopes: List[str] | None = None,
        token_expiry_date: AirbyteDateTime | None = None,
        token_expiry_date_format: str | None = None,
        access_token_name: str = "access_token",
        expires_in_name: str = "expires_in",
        refresh_request_body: Mapping[str, Any] | None = None,
        refresh_request_headers: Mapping[str, Any] | None = None,
        grant_type_name: str = "grant_type",
        grant_type: str = "refresh_token",
        token_expiry_is_time_of_expiration: bool = False,
        refresh_token_error_status_codes: Tuple[int, ...] = (),
        refresh_token_error_key: str = "",
        refresh_token_error_values: Tuple[str, ...] = (),
    ) -> None:
        self._token_refresh_endpoint = token_refresh_endpoint
        self._client_secret_name = client_secret_name
        self._client_secret = client_secret
        self._client_id_name = client_id_name
        self._client_id = client_id
        self._refresh_token_name = refresh_token_name
        self._refresh_token = refresh_token
        self._scopes = scopes
        self._access_token_name = access_token_name
        self._expires_in_name = expires_in_name
        self._refresh_request_body = refresh_request_body
        self._refresh_request_headers = refresh_request_headers
        self._grant_type_name = grant_type_name
        self._grant_type = grant_type

        self._token_expiry_date = token_expiry_date or (ab_datetime_now() - timedelta(days=1))
        self._token_expiry_date_format = token_expiry_date_format
        self._token_expiry_is_time_of_expiration = token_expiry_is_time_of_expiration
        self._access_token = None
        super().__init__(
            refresh_token_error_status_codes, refresh_token_error_key, refresh_token_error_values
        )

    def get_token_refresh_endpoint(self) -> str:
        return self._token_refresh_endpoint

    def get_client_id_name(self) -> str:
        return self._client_id_name

    def get_client_id(self) -> str:
        return self._client_id

    def get_client_secret_name(self) -> str:
        return self._client_secret_name

    def get_client_secret(self) -> str:
        return self._client_secret

    def get_refresh_token_name(self) -> str:
        return self._refresh_token_name

    def get_refresh_token(self) -> str:
        return self._refresh_token

    def get_access_token_name(self) -> str:
        return self._access_token_name

    def get_scopes(self) -> list[str]:
        return self._scopes  # type: ignore[return-value]

    def get_expires_in_name(self) -> str:
        return self._expires_in_name

    def get_refresh_request_body(self) -> Mapping[str, Any]:
        return self._refresh_request_body  # type: ignore[return-value]

    def get_refresh_request_headers(self) -> Mapping[str, Any]:
        return self._refresh_request_headers  # type: ignore[return-value]

    def get_grant_type_name(self) -> str:
        return self._grant_type_name

    def get_grant_type(self) -> str:
        return self._grant_type

    def get_token_expiry_date(self) -> AirbyteDateTime:
        return self._token_expiry_date

    def set_token_expiry_date(self, value: AirbyteDateTime) -> None:
        self._token_expiry_date = value

    @property
    def token_expiry_is_time_of_expiration(self) -> bool:
        return self._token_expiry_is_time_of_expiration

    @property
    def token_expiry_date_format(self) -> Optional[str]:
        return self._token_expiry_date_format

    @property
    def access_token(self) -> str:
        return self._access_token  # type: ignore[return-value]

    @access_token.setter
    def access_token(self, value: str) -> None:
        self._access_token = value  # type: ignore[assignment]  # Incorrect type for assignment


class SingleUseRefreshTokenOauth2Authenticator(Oauth2Authenticator):
    """
    Authenticator that should be used for API implementing single use refresh tokens:
    when refreshing access token some API returns a new refresh token that needs to used in the next refresh flow.
    This authenticator updates the configuration with new refresh token by emitting Airbyte control message from an observed mutation.
    By default, this authenticator expects a connector config with a "credentials" field with the following nested fields: client_id,
    client_secret, refresh_token. This behavior can be changed by defining custom config path (using dpath paths) in client_id_config_path,
    client_secret_config_path, refresh_token_config_path constructor arguments.
    """

    def __init__(
        self,
        connector_config: Mapping[str, Any],
        token_refresh_endpoint: str,
        scopes: List[str] | None = None,
        access_token_name: str = "access_token",
        expires_in_name: str = "expires_in",
        refresh_token_name: str = "refresh_token",
        refresh_request_body: Mapping[str, Any] | None = None,
        refresh_request_headers: Mapping[str, Any] | None = None,
        grant_type_name: str = "grant_type",
        grant_type: str = "refresh_token",
        client_id_name: str = "client_id",
        client_id: Optional[str] = None,
        client_secret_name: str = "client_secret",
        client_secret: Optional[str] = None,
        access_token_config_path: Sequence[str] = ("credentials", "access_token"),
        refresh_token_config_path: Sequence[str] = ("credentials", "refresh_token"),
        token_expiry_date_config_path: Sequence[str] = ("credentials", "token_expiry_date"),
        token_expiry_date_format: Optional[str] = None,
        message_repository: MessageRepository = NoopMessageRepository(),
        token_expiry_is_time_of_expiration: bool = False,
        refresh_token_error_status_codes: Tuple[int, ...] = (),
        refresh_token_error_key: str = "",
        refresh_token_error_values: Tuple[str, ...] = (),
    ) -> None:
        """
        Args:
            connector_config (Mapping[str, Any]): The full connector configuration
            token_refresh_endpoint (str): Full URL to the token refresh endpoint
            scopes (List[str], optional): List of OAuth scopes to pass in the refresh token request body. Defaults to None.
            access_token_name (str, optional): Name of the access token field, used to parse the refresh token response. Defaults to "access_token".
            expires_in_name (str, optional): Name of the name of the field that characterizes when the current access token will expire, used to parse the refresh token response. Defaults to "expires_in".
            refresh_token_name (str, optional): Name of the name of the refresh token field, used to parse the refresh token response. Defaults to "refresh_token".
            refresh_request_body (Mapping[str, Any], optional): Custom key value pair that will be added to the refresh token request body. Defaults to None.
            refresh_request_headers (Mapping[str, Any], optional): Custom key value pair that will be added to the refresh token request headers. Defaults to None.
            grant_type (str, optional): OAuth grant type. Defaults to "refresh_token".
            client_id (Optional[str]): The client id to authenticate. If not specified, defaults to credentials.client_id in the config object.
            client_secret (Optional[str]): The client secret to authenticate. If not specified, defaults to credentials.client_secret in the config object.
            access_token_config_path (Sequence[str]): Dpath to the access_token field in the connector configuration. Defaults to ("credentials", "access_token").
            refresh_token_config_path (Sequence[str]): Dpath to the refresh_token field in the connector configuration. Defaults to ("credentials", "refresh_token").
            token_expiry_date_config_path (Sequence[str]): Dpath to the token_expiry_date field in the connector configuration. Defaults to ("credentials", "token_expiry_date").
            token_expiry_date_format (Optional[str]): Date format of the token expiry date field (set by expires_in_name). If not specified the token expiry date is interpreted as number of seconds until expiration.
            token_expiry_is_time_of_expiration bool: set True it if expires_in is returned as time of expiration instead of the number seconds until expiration
            message_repository (MessageRepository): the message repository used to emit logs on HTTP requests and control message on config update
        """
        self._connector_config = connector_config
        self._client_id: str = self._get_config_value_by_path(
            ("credentials", "client_id"), client_id
        )
        self._client_secret: str = self._get_config_value_by_path(
            ("credentials", "client_secret"), client_secret
        )
        self._client_id_name = client_id_name
        self._client_secret_name = client_secret_name
        self._access_token_config_path = access_token_config_path
        self._refresh_token_config_path = refresh_token_config_path
        self._token_expiry_date_config_path = token_expiry_date_config_path
        self._token_expiry_date_format = token_expiry_date_format
        self._refresh_token_name = refresh_token_name
        self._grant_type_name = grant_type_name
        self._connector_config = connector_config
        self.__message_repository = message_repository
        super().__init__(
            token_refresh_endpoint=token_refresh_endpoint,
            client_id_name=self._client_id_name,
            client_id=self._client_id,
            client_secret_name=self._client_secret_name,
            client_secret=self._client_secret,
            refresh_token=self.get_refresh_token(),
            refresh_token_name=self._refresh_token_name,
            scopes=scopes,
            token_expiry_date=self.get_token_expiry_date(),
            access_token_name=access_token_name,
            expires_in_name=expires_in_name,
            refresh_request_body=refresh_request_body,
            refresh_request_headers=refresh_request_headers,
            grant_type_name=self._grant_type_name,
            grant_type=grant_type,
            token_expiry_date_format=token_expiry_date_format,
            token_expiry_is_time_of_expiration=token_expiry_is_time_of_expiration,
            refresh_token_error_status_codes=refresh_token_error_status_codes,
            refresh_token_error_key=refresh_token_error_key,
            refresh_token_error_values=refresh_token_error_values,
        )

    @property
    def access_token(self) -> str:
        """
        Retrieve the access token from the configuration.

        Returns:
            str: The access token.
        """
        return self._get_config_value_by_path(self._access_token_config_path)  # type: ignore[return-value]

    @access_token.setter
    def access_token(self, new_access_token: str) -> None:
        """
        Sets a new access token.

        Args:
            new_access_token (str): The new access token to be set.
        """
        self._set_config_value_by_path(self._access_token_config_path, new_access_token)

    def get_refresh_token(self) -> str:
        """
        Retrieve the refresh token from the configuration.

        This method fetches the refresh token using the configuration path specified
        by `_refresh_token_config_path`.

        Returns:
            str: The refresh token as a string.
        """
        return self._get_config_value_by_path(self._refresh_token_config_path)  # type: ignore[return-value]

    def set_refresh_token(self, new_refresh_token: str) -> None:
        """
        Updates the refresh token in the configuration.

        Args:
            new_refresh_token (str): The new refresh token to be set.
        """
        self._set_config_value_by_path(self._refresh_token_config_path, new_refresh_token)

    def get_token_expiry_date(self) -> AirbyteDateTime:
        """
        Retrieves the token expiry date from the configuration.

        This method fetches the token expiry date from the configuration using the specified path.
        If the expiry date is an empty string, it returns the current date and time minus one day.
        Otherwise, it parses the expiry date string into an AirbyteDateTime object.

        Returns:
            AirbyteDateTime: The parsed or calculated token expiry date.

        Raises:
            TypeError: If the result is not an instance of AirbyteDateTime.
        """
        expiry_date = self._get_config_value_by_path(self._token_expiry_date_config_path)
        result = (
            ab_datetime_now() - timedelta(days=1)
            if expiry_date == ""
            else ab_datetime_parse(str(expiry_date))
        )
        if isinstance(result, AirbyteDateTime):
            return result
        raise TypeError("Invalid datetime conversion")

    def set_token_expiry_date(self, new_token_expiry_date: AirbyteDateTime) -> None:  # type: ignore[override]
        """
        Sets the token expiry date in the configuration.

        Args:
            new_token_expiry_date (AirbyteDateTime): The new expiry date for the token.
        """
        self._set_config_value_by_path(
            self._token_expiry_date_config_path, str(new_token_expiry_date)
        )

    def token_has_expired(self) -> bool:
        """Returns True if the token is expired"""
        return ab_datetime_now() > self.get_token_expiry_date()

    def get_access_token(self) -> str:
        """Retrieve new access and refresh token if the access token has expired.
        The new refresh token is persisted with the set_refresh_token function
        Returns:
            str: The current access_token, updated if it was previously expired.
        """
        if self.token_has_expired():
            new_access_token, access_token_expires_in, new_refresh_token = (
                self.refresh_access_token()
            )
            self.access_token = new_access_token
            self.set_refresh_token(new_refresh_token)
            self.set_token_expiry_date(access_token_expires_in)
            self._emit_control_message()
        return self.access_token

    def refresh_access_token(self) -> Tuple[str, AirbyteDateTime, str]:  # type: ignore[override]
        """
        Refreshes the access token by making a handled request and extracting the necessary token information.

        Returns:
            Tuple[str, str, str]: A tuple containing the new access token, token expiry date, and refresh token.
        """
        response_json = self._make_handled_request()
        return (
            self._extract_access_token(response_json),
            self._extract_token_expiry_date(response_json),
            self._extract_refresh_token(response_json),
        )

    def _set_config_value_by_path(self, config_path: Union[str, Sequence[str]], value: Any) -> None:
        """
        Set a value in the connector configuration at the specified path.

        Args:
            config_path (Union[str, Sequence[str]]): The path within the configuration where the value should be set.
                This can be a string representing a single key or a sequence of strings representing a nested path.
            value (Any): The value to set at the specified path in the configuration.

        Returns:
            None
        """
        dpath.new(self._connector_config, config_path, value)  # type: ignore[arg-type]

    def _get_config_value_by_path(
        self, config_path: Union[str, Sequence[str]], default: Optional[str] = None
    ) -> str | Any:
        """
        Retrieve a value from the connector configuration using a specified path.

        Args:
            config_path (Union[str, Sequence[str]]): The path to the desired configuration value. This can be a string or a sequence of strings.
            default (Optional[str], optional): The default value to return if the specified path does not exist in the configuration. Defaults to None.

        Returns:
            Any: The value from the configuration at the specified path, or the default value if the path does not exist.
        """
        return dpath.get(
            self._connector_config,  # type: ignore[arg-type]
            config_path,
            default=default if default is not None else "",
        )

    def _emit_control_message(self) -> None:
        """
        Emits a control message based on the connector configuration.

        This method checks if the message repository is not a NoopMessageRepository.
        If it is not, it emits a message using the message repository. Otherwise,
        it falls back to emitting the configuration as an Airbyte control message
        directly to the console for backward compatibility.

        Note:
            The function `emit_configuration_as_airbyte_control_message` has been deprecated
            in favor of the package `airbyte_cdk.sources.message`.

        Raises:
            TypeError: If the argument types are incorrect.
        """
        # FIXME emit_configuration_as_airbyte_control_message as been deprecated in favor of package airbyte_cdk.sources.message
        # Usually, a class shouldn't care about the implementation details but to keep backward compatibility where we print the
        # message directly in the console, this is needed
        if not isinstance(self._message_repository, NoopMessageRepository):
            self._message_repository.emit_message(
                create_connector_config_control_message(self._connector_config)  # type: ignore[arg-type]
            )
        else:
            emit_configuration_as_airbyte_control_message(self._connector_config)  # type: ignore[arg-type]

    @property
    def _message_repository(self) -> MessageRepository:
        """
        Overriding AbstractOauth2Authenticator._message_repository to allow for HTTP request logs
        """
        return self.__message_repository



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass, field
from typing import Any, Iterable, List, Mapping, Optional, Union

import requests

from airbyte_cdk.sources.declarative.extractors.http_selector import HttpSelector
from airbyte_cdk.sources.declarative.extractors.record_extractor import RecordExtractor
from airbyte_cdk.sources.declarative.extractors.record_filter import RecordFilter
from airbyte_cdk.sources.declarative.extractors.type_transformer import (
    TypeTransformer as DeclarativeTypeTransformer,
)
from airbyte_cdk.sources.declarative.interpolation import InterpolatedString
from airbyte_cdk.sources.declarative.models import SchemaNormalization
from airbyte_cdk.sources.declarative.retrievers.file_uploader import FileUploader
from airbyte_cdk.sources.declarative.transformations import RecordTransformation
from airbyte_cdk.sources.types import Config, Record, StreamSlice, StreamState
from airbyte_cdk.sources.utils.transform import TypeTransformer


@dataclass
class RecordSelector(HttpSelector):
    """
    Responsible for translating an HTTP response into a list of records by extracting records from the response and optionally filtering
    records based on a heuristic.

    Attributes:
        extractor (RecordExtractor): The record extractor responsible for extracting records from a response
        schema_normalization (TypeTransformer): The record normalizer responsible for casting record values to stream schema types
        record_filter (RecordFilter): The record filter responsible for filtering extracted records
        transformations (List[RecordTransformation]): The transformations to be done on the records
    """

    extractor: RecordExtractor
    config: Config
    parameters: InitVar[Mapping[str, Any]]
    schema_normalization: Union[TypeTransformer, DeclarativeTypeTransformer]
    name: str
    _name: Union[InterpolatedString, str] = field(init=False, repr=False, default="")
    record_filter: Optional[RecordFilter] = None
    transformations: List[RecordTransformation] = field(default_factory=lambda: [])
    transform_before_filtering: bool = False
    file_uploader: Optional[FileUploader] = None

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        self._parameters = parameters
        self._name = (
            InterpolatedString(self._name, parameters=parameters)
            if isinstance(self._name, str)
            else self._name
        )

    @property  # type: ignore
    def name(self) -> str:
        """
        :return: Stream name
        """
        return (
            str(self._name.eval(self.config))
            if isinstance(self._name, InterpolatedString)
            else self._name
        )

    @name.setter
    def name(self, value: str) -> None:
        if not isinstance(value, property):
            self._name = value

    def select_records(
        self,
        response: requests.Response,
        stream_state: StreamState,
        records_schema: Mapping[str, Any],
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Iterable[Record]:
        """
        Selects records from the response
        :param response: The response to select the records from
        :param stream_state: The stream state
        :param records_schema: json schema of records to return
        :param stream_slice: The stream slice
        :param next_page_token: The paginator token
        :return: List of Records selected from the response
        """
        all_data: Iterable[Mapping[str, Any]] = self.extractor.extract_records(response)
        yield from self.filter_and_transform(
            all_data, stream_state, records_schema, stream_slice, next_page_token
        )

    def filter_and_transform(
        self,
        all_data: Iterable[Mapping[str, Any]],
        stream_state: StreamState,
        records_schema: Mapping[str, Any],
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Iterable[Record]:
        """
        There is an issue with the selector as of 2024-08-30: it does technology-agnostic processing like filtering, transformation and
        normalization with an API that is technology-specific (as requests.Response is only for HTTP communication using the requests
        library).

        Until we decide to move this logic away from the selector, we made this method public so that users like AsyncJobRetriever could
        share the logic of doing transformations on a set of records.
        """
        if self.transform_before_filtering:
            transformed_data = self._transform(all_data, stream_state, stream_slice)
            transformed_filtered_data = self._filter(
                transformed_data, stream_state, stream_slice, next_page_token
            )
        else:
            filtered_data = self._filter(all_data, stream_state, stream_slice, next_page_token)
            transformed_filtered_data = self._transform(filtered_data, stream_state, stream_slice)
        normalized_data = self._normalize_by_schema(
            transformed_filtered_data, schema=records_schema
        )
        for data in normalized_data:
            record = Record(data=data, stream_name=self.name, associated_slice=stream_slice)
            if self.file_uploader:
                self.file_uploader.upload(record)
            yield record

    def _normalize_by_schema(
        self, records: Iterable[Mapping[str, Any]], schema: Optional[Mapping[str, Any]]
    ) -> Iterable[Mapping[str, Any]]:
        if schema:
            # record has type Mapping[str, Any], but dict[str, Any] expected
            for record in records:
                normalized_record = dict(record)
                self.schema_normalization.transform(normalized_record, schema)
                yield normalized_record
        else:
            yield from records

    def _filter(
        self,
        records: Iterable[Mapping[str, Any]],
        stream_state: StreamState,
        stream_slice: Optional[StreamSlice],
        next_page_token: Optional[Mapping[str, Any]],
    ) -> Iterable[Mapping[str, Any]]:
        if self.record_filter:
            yield from self.record_filter.filter_records(
                records,
                stream_state=stream_state,
                stream_slice=stream_slice,
                next_page_token=next_page_token,
            )
        else:
            yield from records

    def _transform(
        self,
        records: Iterable[Mapping[str, Any]],
        stream_state: StreamState,
        stream_slice: Optional[StreamSlice] = None,
    ) -> Iterable[Mapping[str, Any]]:
        for record in records:
            for transformation in self.transformations:
                transformation.transform(
                    record,  # type: ignore  # record has type Mapping[str, Any], but Dict[str, Any] expected
                    config=self.config,
                    stream_state=stream_state,
                    stream_slice=stream_slice,
                )
            yield record



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


class CircularReferenceException(Exception):
    """
    Raised when a circular reference is detected in a manifest.
    """

    def __init__(self, reference: str) -> None:
        super().__init__(f"Circular reference found: {reference}")


class UndefinedReferenceException(Exception):
    """
    Raised when refering to an undefined reference.
    """

    def __init__(self, path: str, reference: str) -> None:
        super().__init__(f"Undefined reference {reference} from {path}")
